﻿#!/usr/bin/python
# -*- coding: utf-8 -*-
#
#  exoMaker.py - make a slideshow using AviUtl video tool.
#
#  (c)2018 Teruyoshi Fujiwara <tf@dsl.gr.jp> All Rights Reserved
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import sys
import codecs
import json
import re

class ExEdit:
    def __init__(self):
        # Variables defined in the EXO format.
        self.width = 640
        self.height = 480
        self.rate = 30
        self.scale = 1
        self.length = -1
        self.audio_rate = 44100
        self.audio_ch = 2

        # other variables
        self.imageList = []
        self.font_size = 16
        self.fade_time_offset = 5
        self.movie_volume = 50

    def readConfig(self, file):
        with open(file) as f:
            js = json.load(f)
            self.width = js['width']
            self.height = js['height']
            self.rate = js['rate']
            self.scale = js['scale']
            self.audio_rate = js['audio_rate']
            self.audio_ch = js['audio_ch']
            
            self.font_size = js['font_size']
            self.fade_time_offset = js['fade_time_offset']
            self.movie_volume = js['movie_volume']

    def readScript(self, file):
        fin  = codecs.open(file, 'r', 'shift_jis')
        self.imageList = Parser.parse(self, fin)

    def print(self, file):
        f = codecs.open(file, 'w', 'shift_jis')

        self.updateLength()
        
        f.write('[exedit]\r\n')
        f.write('width={}\r\n'.format(self.width))
        f.write('height={}\r\n'.format(self.height))
        f.write('rate={}\r\n'.format(self.rate))
        f.write('scale={}\r\n'.format(self.scale))
        f.write('length={}\r\n'.format(self.length))
        f.write('audio_rate={}\r\n'.format(self.audio_rate))
        f.write('audio_ch={}\r\n'.format(self.audio_ch))

        obj_num = 0

        # Output Image
        start = 1
        for img in self.imageList:
            end = start + img.frames - 1
            img.start = start
            img.end = end
            if img.file_path != '':
                img.print(f, obj_num)
                obj_num = obj_num + 1
            start = end + 1

        # Output SceneChange
        for img in self.imageList:
            start = img.start - self.fade_time_offset
            end = img.start + self.fade_time_offset
            type = img.trans
            if start > 0 and img.trans >= 0:
                SceneChange.print(self, f, obj_num, start, end, type)
                obj_num = obj_num + 1

        # Output Text
        for img in self.imageList:
            start = img.start
            end = img.end
            if img.caption != '':
                if img.font_size > 0:
                    fs = img.font_size
                else:
                    fs = self.font_size
                Caption.print(self, f, obj_num, start + self.fade_time_offset, end, fs, img.caption, img.alignment)
                obj_num = obj_num + 1

        # Output Audio linked with movie
        for img in self.imageList:
            start = img.start
            end = img.end
            if re.search(r'\.mp4$', img.file_path, re.IGNORECASE):
                AudioObj.print(self, f, obj_num, start, end, self.movie_volume)
                obj_num = obj_num + 1

    def updateLength(self):
        self.length = 0
        for img in self.imageList:
            self.length = self.length + img.frames

class ImageObj:
    def __init__(self, file_path, frames, trans, caption, font_size, alignment):
        self.file_path = file_path
        self.frames = frames
        self.trans = trans
        self.caption = caption
        self.start = 0
        self.end = 0

        self.caption = self.caption.replace('\\n', '\n')
        self.font_size = font_size
        self.alignment = alignment
        if alignment == '':
            self.alignment = 'B'
        else:
            self.alignment = alignment

    def print(self, f, num):
        if re.search(r'\.jpg$', self.file_path, re.IGNORECASE):
            self.print_jpg(f, num)
        elif re.search(r'\.mp4$', self.file_path, re.IGNORECASE):
            self.print_mp4(f, num)
        else:
            sys.stderr.write('Unknown file type:{}\r\n'.format(self.file_path))
            exit()
        
    def print_jpg(self, f, num):
        f.write('[{}]\r\n'.format(num))
        f.write('start={}\r\n'.format(self.start))
        f.write('end={}\r\n'.format(self.end))
        f.write('layer=1\r\n')
        f.write('overlay=1\r\n')
        f.write('camera=0\r\n')
        f.write('[{}.0]\r\n'.format(num))
        f.write('_name=画像ファイル\r\n')
        f.write('file={}\r\n'.format(self.file_path))
        f.write('[{}.1]\r\n'.format(num))
        f.write('_name=標準描画\r\n')
        f.write('X=0.0\r\n')
        f.write('Y=0.0\r\n')
        f.write('Z=0.0\r\n')
        f.write('拡大率=100.00\r\n')
        f.write('透明度=0.0\r\n')
        f.write('回転=0.00\r\n')
        f.write('blend=0\r\n')

    def print_mp4(self, f, num):
        f.write('[{}]\r\n'.format(num))
        f.write('start={}\r\n'.format(self.start))
        f.write('end={}\r\n'.format(self.end))
        f.write('layer=1\r\n')
        f.write('group=1\r\n')
        f.write('overlay=1\r\n')
        f.write('camera=0\r\n')
        f.write('[{}.0]\r\n'.format(num))
        f.write('_name=動画ファイル\r\n')
        f.write('再生位置=1\r\n')
        f.write('再生速度=100.0\r\n')
        f.write('ループ再生=0\r\n')
        f.write('アルファチャンネルを読み込む=0\r\n')
        f.write('file={}\r\n'.format(self.file_path))
        f.write('[{}.1]\r\n'.format(num))
        f.write('_name=標準描画\r\n')
        f.write('X=0.0\r\n')
        f.write('Y=0.0\r\n')
        f.write('Z=0.0\r\n')
        f.write('拡大率=100.00\r\n')
        f.write('透明度=0.0\r\n')
        f.write('回転=0.00\r\n')
        f.write('blend=0\r\n')

class SceneChange:
    def print(self, f, num, start, end, type):
        f.write('[{}]\r\n'.format(num))
        f.write('start={}\r\n'.format(start))
        f.write('end={}\r\n'.format(end))
        f.write('layer=2\r\n')
        f.write('[{}.0]\r\n'.format(num))
        f.write('_name=シーンチェンジ\r\n')
        f.write('調整=0.00\r\n')
        f.write('track1=0.00\r\n')
        f.write('反転=0\r\n')
        f.write('check0=0\r\n')
        f.write('type={}\r\n'.format(type))
        f.write('filter=0\r\n')
        f.write('name=\r\n')
        f.write('param=*\r\n')

class Caption:
    FONT_Y_OFFSET_B = 330
    FONT_Y_OFFSET_C = -50
    
    def print(self, f, num, start, end, font_size, text, align):
        if align == 'B':
            align_num = 7
            y_offset = Caption.FONT_Y_OFFSET_B
        elif align == 'C':
            align_num = 4
            y_offset = Caption.FONT_Y_OFFSET_C
        else:
            print('Unknown Alignment:{}'.format(align))
        f.write('[{}]\r\n'.format(num))
        f.write('start={}\r\n'.format(start))
        f.write('end={}\r\n'.format(end))
        f.write('layer=3\r\n')
        f.write('overlay=1\r\n')
        f.write('camera=0\r\n')
        f.write('[{}.0]\r\n'.format(num))
        f.write('_name=テキスト\r\n')
        f.write('サイズ={}\r\n'.format(font_size))
        f.write('表示速度=0.0\r\n')
        f.write('文字毎に個別オブジェクト=0\r\n')
        f.write('移動座標上に表示する=0\r\n')
        f.write('自動スクロール=0\r\n')
        f.write('B=1\r\n')
        f.write('I=0\r\n')
        f.write('type=4\r\n')
        f.write('autoadjust=0\r\n')
        f.write('soft=1\r\n')
        f.write('monospace=0\r\n')
        f.write('align={}\r\n'.format(align_num))
        f.write('spacing_x=0\r\n')
        f.write('spacing_y=0\r\n')
        f.write('precision=1\r\n')
        f.write('color=ffffff\r\n')
        f.write('color2=000080\r\n')
        f.write('font=Meiryo UI\r\n')
        f.write('text={}\r\n'.format(Caption.encodeText(self, text)))
        f.write('[{}.1]\r\n'.format(num))
        f.write('_name=フェード\r\n')
        f.write('イン=0.50\r\n')
        f.write('アウト=0.50\r\n')
        f.write('[{}.2]\r\n'.format(num))
        f.write('_name=標準描画\r\n')
        f.write('X=0\r\n')
        f.write('Y={}\r\n'.format(y_offset))
        f.write('Z=0.0\r\n')
        f.write('拡大率=100.00\r\n')
        f.write('透明度=0.0\r\n')
        f.write('回転=0.00\r\n')
        f.write('blend=0\r\n')
        
    def encodeText(self, text):
        # Unicode でエンコードした後，合計が4096バイトとなるように '0'で埋める．
        hex = codecs.encode(text.encode('utf-16le'), 'hex_codec')
        result = hex.decode() + '0' * (4096 - len(hex))
        return result

class AudioObj:
    def print(self, f, num, start, end, movie_volume):
        f.write('[{}]\r\n'.format(num))
        f.write('start={}\r\n'.format(start))
        f.write('end={}\r\n'.format(end))
        f.write('layer=4\r\n')
        f.write('group=1\r\n')
        f.write('overlay=1\r\n')
        f.write('audio=1\r\n')
        f.write('[{}.0]\r\n'.format(num))
        f.write('_name=音声ファイル\r\n')
        f.write('再生位置=0.00\r\n')
        f.write('再生速度=100.0\r\n')
        f.write('ループ再生=0\r\n')
        f.write('動画ファイルと連携=1\r\n')
        f.write('file=\r\n')
        f.write('[{}.1]\r\n'.format(num))
        f.write('_name=音量フェード\r\n')
        f.write('イン=0.50\r\n')
        f.write('アウト=0.50\r\n')
        f.write('[{}.2]\r\n'.format(num))
        f.write('_name=標準再生\r\n')
        f.write('音量={}\r\n'.format(movie_volume))
        f.write('左右=0.0\r\n')

class Parser:
    def parse(self, file):
        list = []
        for line in file:
            line = line.strip()
            cols = line.split(',')
            if cols[0] == 'image':
                if cols[5] == '':
                    font_size = -1
                else:
                    font_size = int(cols[5])
                img = ImageObj(cols[1], int(cols[2]), int(cols[3]), cols[4], font_size, cols[6])
                list.append(img)
            else:
                print('Unknowen statement.'.format(cols[0]))
        return list

def main():
    exedit = ExEdit()

    exedit.readConfig(sys.argv[1])
    exedit.readScript(sys.argv[2])
    exedit.print(sys.argv[3])

if __name__ == "__main__":
    main()
